Attribute VB_Name = "SpeedTest"
Option Explicit

Private Declare Function QueryPerformanceCounter Lib "kernel32" (lpPerformanceCount As Currency) As Long
Private Declare Function QueryPerformanceFrequency Lib "kernel32" (lpFrequency As Currency) As Long

Private CPU_Takt As Currency, Overhead As Currency

Private Const k_msec As Currency = 1000
Private Const k_usec As Currency = 1000000
Private Const k_nsec As Currency = 1000000000

Public Sub Speed_Test()
  Dim block1() As Byte, i As Long, block2(3, 3) As Byte
  Dim bkey(15) As Byte, y As Currency, c1 As Currency
  Dim c2 As Currency, clk As New cCpuClk
  
    Const rnds As Long = 1000
  
    Randomize Timer
    Init_Speedtest
    ReDim block1(3, 3)
    
    y = 0
    Rijndael_EncryptBlock block1(0, 0), key(0, 0), t(0, 0), 0&
    For i = 1 To rnds
      Random_Key bkey
      KeyExpansion bkey, 128
      Set_Random block1
      
      clk.CpuClk c1
      Rijndael_EncryptBlock block1(0, 0), key(0, 0), t(0, 0), 0&
      clk.CpuClk c2
      y = y + (c2 - c1 - Overhead)
    Next i
    y = y / rnds
    
    Log "Bentigte Zeit: " & CStr(nsec(y)) & " ns (Verschlsselung)"
    Log "Durchsatz: " & Format$(CStr(Round(16 / sec(y), 2)), "###,###,###.00") & " B/s"
    
    y = 0
    Rijndael_DecryptBlock block1(0, 0), ikey(0, 0), t2(0, 0), 0&
    For i = 1 To rnds
      Random_Key bkey
      InvKeyExpansion bkey, 128
      Set_Random block1
      
      clk.CpuClk c1
      Rijndael_DecryptBlock block1(0, 0), ikey(0, 0), t2(0, 0), 0&
      clk.CpuClk c2
      y = y + (c2 - c1 - Overhead)
    Next i
    y = y / rnds
    
    Log "Bentigte Zeit: " & CStr(nsec(y)) & " ns (Entschlsselung)"
    Log "Durchsatz: " & Format$(CStr(Round(16 / sec(y), 2)), "###,###,###.00") & " B/s"
End Sub

Public Sub Init_Speedtest()
  Dim cCycles As Currency, cStart As Currency, cStop As Currency, cNow As Currency
  Dim c1 As Currency, c2 As Currency, clk As New cCpuClk, Factor As Currency, i As Long

    For i = 1 To 1000
      clk.CpuClk c1
      clk.CpuClk c2
    Next i
    Overhead = c2 - c1
    
    Call QueryPerformanceFrequency(cCycles)
    
    Call clk.CpuClk(c1)
      Call QueryPerformanceCounter(cStart)
      cStop = cStart + cCycles
    
      Do
        Call QueryPerformanceCounter(cNow)
      Loop Until (cNow >= cStop)
    Call clk.CpuClk(c2)
    c2 = c2 - c1 - Overhead
    Factor = (cNow - cStart) / cCycles
    CPU_Takt = c2 / Factor
End Sub

Public Function sec(ByVal zeit As Currency) As Double
    sec = zeit / CPU_Takt
End Function

Public Function msec(ByVal zeit As Currency) As Double
    msec = zeit / CPU_Takt * k_msec
End Function

Public Function usec(ByVal zeit As Currency) As Double
    usec = zeit / CPU_Takt * k_usec
End Function

Public Function nsec(ByVal zeit As Currency) As Double
    nsec = zeit / CPU_Takt * k_nsec
End Function

Private Sub Set_Random(block() As Byte)
  Dim i As Long, n As Long
  
    ReDim block(3, 3)
  
    For i = 0 To 3
      For n = 0 To 3
        block(i, n) = Int(Rnd * 255)
      Next n
    Next i
End Sub

Private Sub Random_Key(k() As Byte)
  Dim i As Long
  
    For i = 0 To UBound(k)
      k(i) = Int(Rnd * 255)
    Next i
End Sub

Private Sub Copy_Block(block() As Byte, block2() As Byte)
  Dim i As Long, n As Long
    
    For i = 0 To 3
      For n = 0 To 3
        block2(i, n) = block(i, n)
      Next n
    Next i
End Sub

Private Sub Check_Block(block() As Byte, block2() As Byte)
  Dim i As Long, n As Long
  
    For i = 0 To 3
      For n = 0 To 3
        If block(i, n) <> block2(i, n) Then
          Log "Fehler aufgetreten: Speed_Test"
          PrintBlock block
          PrintBlock block2
        End If
      Next n
    Next i
End Sub
